/****** amxmouse.acepansion/--background-- **********************************
*
* DESCRIPTION
*  This plugin provides AMX Mouse emulation to ACE CPC Emulator.
*
* HISTORY
*
*****************************************************************************
*
*/

#include <proto/exec.h>
#include <proto/utility.h>
#include <proto/dos.h>

#include <exec/interrupts.h>

#ifndef __HAIKU__
#include <devices/input.h>
#include <devices/inputevent.h>
#endif

#include <devices/rawkeycodes.h>
#include <clib/alib_protos.h>

#include <acepansion/lib_header.h>
#include <libraries/acepansion_plugin.h>

#include "acepansion.h"
#define CATCOMP_NUMBERS
#include "generated/locale_strings.h" /* catalog string ids */

#ifdef __HAIKU__
#include <Application.h>
#include <InterfaceDefs.h>
#include <Point.h>
#include <View.h>
#include <Window.h>
#endif

struct PluginData
{
    // Plugin generic part (do NEVER put anything before this!)
    struct ACEpansionPlugin common;

    // AMX Mouse specifics
    BYTE mouseDeltaX;
    BYTE mouseDeltaY;
    BOOL mouseActive;
    BOOL mouseLeft;
    BOOL mouseRight;
    BOOL mouseMiddle;
    BYTE inputPri;
    struct IOStdReq  *InputIOReq;
    struct MsgPort   *InputMsgPort;
    struct Interrupt *InputHandler;
};



#ifndef __HAIKU__
struct Library *DOSBase;
struct Library *UtilityBase;

static IPTR inputhandler(VOID);
static struct EmulLibEntry ihgate = { TRAP_LIB, 0, (void (*)(void))inputhandler };

#define inputhandler_code (void (*)(void))&ihgate
#define inputhandler_name "AMX Mouse ACEpansion"



/*
** Private function which is called to initialize commons
** just after the library was loaded into memory and prior
** to any other API call.
**
** This is the good place to open our required libraries.
*/
BOOL InitResources(VOID)
{
    DOSBase = OpenLibrary("dos.library", 0L);
    UtilityBase = OpenLibrary("utility.library", 0L);

    return (DOSBase && UtilityBase);
}



/*
** Private function which is called to free commons
** just before the library is expurged from memory
**
** This is the good place to close our required libraries.
*/
VOID FreeResources(VOID)
{
    CloseLibrary(UtilityBase);
    UtilityBase = NULL;

    CloseLibrary(DOSBase);
    DOSBase = NULL;
}
#endif


#ifdef __HAIKU__
extern "C" int ToggleMouse()
{
    static int32 oldModifiers;

    if (oldModifiers != 0) {
        oldModifiers = modifiers();
        return 0;
    }
    oldModifiers = modifiers();
    return (modifiers() & B_RIGHT_CONTROL_KEY) != 0;
}


extern "C" void GetMouseState(BYTE* x, BYTE* y, BOOL* left, BOOL* right)
{
    static BPoint oldPoint;
    BPoint point;
    uint32 buttons;
    get_mouse(&point, &buttons);

    *left = (buttons & 1);
    *right = (buttons & 2);
    *x += (point.x - oldPoint.x) / 2;
    *y += (oldPoint.y - point.y) / 2;

    oldPoint = point;
}
#endif

/****** amxmouse.acepansion/CreatePlugin ************************************
*
* NAME
*   CreatePlugin -- (V1)
*
* SYNOPSIS
*   struct ACEpansionPlugin * CreatePlugin(CONST_STRPTR *toolTypes, struct SignalSemaphore *sema)
*
* FUNCTION
*   Create the plugin data structure and configure how it will be handled by
*   ACE. This function is usually called at ACE startup.
*
* INPUTS
*   *toolType
*     A pointer to a NULL terminated array of CONST_STRPTR containing the
*     tooltypes provided with the plugin library.
*     The ACTIVE tooltype is handled internally by ACE for all plugins, but
*     you can here handle additional tooltypes to configure your plugin.
*   *sema
*     Pointer to a semaphore to use to protect access to custom data from
*     ACEpansionPlugin when accessed from subtasks or within your PrefsWindow.
*
* RESULT
*   A pointer to an allocated and initialized plugin data structure.
*
* SEE ALSO
*   DeletePlugin
*   ActivatePlugin
*
*****************************************************************************
*
*/

struct ACEpansionPlugin * CreatePlugin(CONST_STRPTR *toolTypes, UNUSED struct SignalSemaphore *sema)
{
    // We are allocating an extended plugin structure to add our custom stuff
    struct PluginData *myPlugin = (struct PluginData*)AllocVec(sizeof(struct PluginData), MEMF_PUBLIC|MEMF_CLEAR);

    if(myPlugin)
    {
        BOOL inputHandlerReady = FALSE;

        // Initialize plugin generic part
        myPlugin->common.ap_APIVersion     = API_VERSION;
        myPlugin->common.ap_Flags          = ACE_FLAGSF_ACTIVE_JOYSTICK | ACE_FLAGSF_EXCLUSIVE_MOUSE;
        myPlugin->common.ap_Title          = GetString(MSG_TITLE);
        myPlugin->common.ap_HelpFileName   = LIBNAME".guide";
        myPlugin->common.ap_ToggleMenuName = GetString(MSG_MENU_TOGGLE);
        myPlugin->common.ap_PrefsMenuName  = NULL;
        myPlugin->common.ap_PrefsWindow    = NULL;

#ifdef __HAIKU__
        inputHandlerReady = TRUE;
#else
        // Initialize plugin specifics
        myPlugin->inputPri = 120;

        if((myPlugin->InputHandler = AllocVec(sizeof(struct Interrupt), MEMF_PUBLIC|MEMF_CLEAR)))
        {
            if((myPlugin->InputMsgPort = CreateMsgPort()))
            {
                if((myPlugin->InputIOReq = CreateStdIO(myPlugin->InputMsgPort)))
                {
                    if(OpenDevice("input.device", 0, (struct IORequest *)myPlugin->InputIOReq,0) == 0)
                    {
                        inputHandlerReady = TRUE;
                    }
                }
            }
        }
#endif

        if(!inputHandlerReady)
        {
            DeletePlugin((struct ACEpansionPlugin *)myPlugin);
            myPlugin = NULL;
        }
        else
        {
            while(*toolTypes)
            {
                CONST_STRPTR ptr = *toolTypes;
                CONST_STRPTR name = ptr;
                ULONG len = 0;

                if(*ptr != '(')
                {
                    while(*ptr != '=' && *ptr != '\0')
                    {
                        ptr++;
                        len++;
                    }

                    if(*ptr == '=')
                    {
                        if(!Strnicmp(name,"PRIORITY",len))
                        {
                            LONG pri;

                            if(StrToLong(ptr+1,&pri) != -1)
                            {
                                myPlugin->inputPri = pri;
                            }
                        }
                    }
                }                        
                toolTypes++;
            }
        }
    }

    return (struct ACEpansionPlugin *)myPlugin;
}



/****** amxmouse.acepansion/DeletePlugin ************************************
*
* NAME
*   DeletePlugin -- (V1)
*
* SYNOPSIS
*   VOID DeletePlugin(struct ACEpansionPlugin *plugin)
*
* FUNCTION
*   Delete a plugin data structure. This function is usually called when ACE
*   is exiting.
*
* INPUTS
*   *plugin
*     Pointer to the plugin data structure to delete.
*
* RESULT
*
* SEE ALSO
*   CreatePlugin
*   ActivatePlugin
*
*****************************************************************************
*
*/

VOID DeletePlugin(struct ACEpansionPlugin *plugin)
{
    struct PluginData *myPlugin = (struct PluginData *)plugin;

#ifndef __HAIKU__
    if(myPlugin->InputHandler)
    {
        if(myPlugin->InputMsgPort)
        {
            if(myPlugin->InputIOReq)
            {
                CloseDevice((struct IORequest *)myPlugin->InputIOReq);
                DeleteStdIO(myPlugin->InputIOReq);
                myPlugin->InputIOReq = NULL;
            }
            DeleteMsgPort(myPlugin->InputMsgPort);
            myPlugin->InputMsgPort = NULL;
        }
        FreeVec(myPlugin->InputHandler);
        myPlugin->InputHandler = NULL;
    }
#endif

    FreeVec(myPlugin);
}



/****** amxmouse.acepansion/ActivatePlugin **********************************
*
* NAME
*   ActivatePlugin -- (V1)
*
* SYNOPSIS
*   BOOL ActivatePlugin(struct ACEpansionPlugin *plugin, BOOL status)
*
* FUNCTION
*   This function is called everytime ACE wants to activate or disactivate your plugin.
*   While you are disactivated, no other function (except Delete) will the called by ACE.
*   Please note that a plugin is always disactivated by ACE prior to DeletePlugin().
*
* INPUTS
*   activate
*     New activation status requested by ACE.
*
* RESULT
*   Actual activation status (you may fail at activation for some reason, but you
*   should never fail at disactivation request).
*
* SEE ALSO
*   CreatePlugin
*   DeletePlugin
*
*****************************************************************************
*
*/

BOOL ActivatePlugin(UNUSED struct ACEpansionPlugin *plugin, BOOL activate)
{
    struct PluginData *myPlugin = (struct PluginData *)plugin;

    myPlugin->mouseActive = FALSE;
    myPlugin->mouseDeltaX = 0;
    myPlugin->mouseDeltaY = 0;
    myPlugin->mouseLeft   = FALSE;
    myPlugin->mouseRight  = FALSE;
    myPlugin->mouseMiddle = FALSE;

#ifndef __HAIKU__
    if(activate)
    {
        myPlugin->InputHandler->is_Node.ln_Type = NT_INTERRUPT;
        myPlugin->InputHandler->is_Node.ln_Pri  = myPlugin->inputPri; /* Priority in food chain */
        myPlugin->InputHandler->is_Node.ln_Name = inputhandler_name;  /* Name of handler */
        myPlugin->InputHandler->is_Code         = inputhandler_code;  /* Address of code */
        myPlugin->InputHandler->is_Data         = myPlugin;           /* User Value passed in A1 */

        myPlugin->InputIOReq->io_Data    = (APTR)myPlugin->InputHandler; /* Point to the structure */
        myPlugin->InputIOReq->io_Command = IND_ADDHANDLER;               /* ADD command ... */
        DoIO((struct IORequest *)myPlugin->InputIOReq);                  /* Send the command */
    }
    else
    {
        myPlugin->InputIOReq->io_Data    = (APTR)myPlugin->InputHandler; /* Which handler to REM */
        myPlugin->InputIOReq->io_Command = IND_REMHANDLER;               /* REM command */
        DoIO((struct IORequest *)myPlugin->InputIOReq);                  /* Send the command */
    }
#endif

    return activate;
}



/****** amxmouse.acepansion/Reset *******************************************
*
* NAME
*   Reset -- (V1)
*
* SYNOPSIS
*   VOID Reset(struct ACEpansionPlugin *plugin)
*
* FUNCTION
*   Function that is called everytime ACE is issuing a bus reset.
*
* INPUTS
*   *plugin
*     Pointer to the plugin data structure to use.
*
* RESULT
*
* SEE ALSO
*   Emulate
*
*****************************************************************************
*
*/

VOID Reset(UNUSED struct ACEpansionPlugin *plugin)
{
}



/****** amxmouse.acepansion/Emulate *****************************************
*
* NAME
*   Emulate -- (V1)
*
* SYNOPSIS
*   VOID Emulate(struct ACEpansionPlugin *plugin, ULONG *signals)
*
* FUNCTION
*   Function that is called at each microsecond of the emulation step inside ACE.
*   Note: Implement this function only if is it really required for your plugin because it
*   is obviously CPU intensive.
*
* INPUTS
*   *plugin
*     Pointer to the plugin data structure to use.
*   *signals
*     Pointer to the I/O signals from the expansion port you can use or alter depending
*     on what you are emulating.
*     The signals bits can be decoded using the following flags:
*       ACE_SIGNALF_INT
*
* RESULT
*
* SEE ALSO
*   Reset
*
*****************************************************************************
*
*/

VOID Emulate(UNUSED struct ACEpansionPlugin *plugin, UNUSED ULONG *signals)
{
}



/****** amxmouse.acepansion/WriteIO *****************************************
*
* NAME
*   WriteIO -- (V1)
*
* SYNOPSIS
*   VOID WriteIO(struct ACEpansionPlugin *plugin, USHORT port, UBYTE value)
*
* FUNCTION
*   Function that is called everytime Z80 is performing a I/O port write operation.
*   Note: Implement this function only if is it really required for your plugin because it
*   is obviously CPU intensive.
*
* INPUTS
*   port
*     Z80 port on with the I/O write operation was issued.
*   value
*     Value that was issued on the port.
*
* RESULT
*
* SEE ALSO
*   ReadIO
*
*****************************************************************************
*
*/

VOID WriteIO(UNUSED struct ACEpansionPlugin *plugin, UNUSED USHORT port, UNUSED UBYTE value)
{
}



/****** amxmouse.acepansion/ReadIO ******************************************
*
* NAME
*   ReadIO -- (V1)
*
* SYNOPSIS
*   VOID ReadIO(struct ACEpansionPlugin *plugin, USHORT port, UBYTE *value)
*
* FUNCTION
*   Function that is called everytime Z80 is performing a I/O port read operation.
*   Note: Implement this function only if is it really required for your plugin because it
*   is obviously CPU intensive.
*
* INPUTS
*   port
*     Z80 port on with the I/O read operation was issued.
*   *value
*     Value to be returned.
*
* RESULT
*
* SEE ALSO
*   WriteIO
*
*****************************************************************************
*
*/

VOID ReadIO(UNUSED struct ACEpansionPlugin *plugin, UNUSED USHORT port, UNUSED UBYTE *value)
{
}



/****** amxmouse.acepansion/GetAudioSample **********************************
*
* NAME
*   GetAudioSample -- (V1)
*
* SYNOPSIS
*   VOID GetAudio(struct ACEpansionPlugin *plugin, SHORT *leftSample, SHORT *rightSample)
*
* FUNCTION
*   Function to be used when your plugin needs to mix an audio output with the one
*   from ACE.
*   It is called at each PSG emulation step, which means at 125KHz.
*
* INPUTS
*   *leftSample
*     Pointer to store the left audio sample to be mixed with ACE audio.
*   *rightSample
*     Pointer to store the right audio sample to be mixed with ACE audio.
*
* RESULT
*
* SEE ALSO
*
*****************************************************************************
*
*/

VOID GetAudio(UNUSED struct ACEpansionPlugin *plugin, UNUSED SHORT *leftSample, UNUSED SHORT *rightSample)
{
}



/****** amxmouse.acepansion/Printer *****************************************
*
* NAME
*   Printer -- (V1)
*
* SYNOPSIS
*   VOID WritePrinter(struct ACEpansionPlugin *plugin, UBYTE data, BOOL strobe, BOOL *busy)
*
* FUNCTION
*   Function that is called everytime the CPC is accessing the printer port.
*
* INPUTS
*   data
*     8 bits written value (note that CPC is limited to 7 bits values, only Amstrad Plus
*     can actually provide the 8th bit).
*   strobe
*     Status of the /STROBE signal from the printer port.
*   *busy
*     Busy signal value that can be altered.
*
* RESULT
*
* SEE ALSO
*
*****************************************************************************
*
*/

VOID Printer(UNUSED struct ACEpansionPlugin *plugin, UNUSED UBYTE data, UNUSED BOOL strobe, UNUSED BOOL *busy)
{
}



/****** amxmouse.acepansion/Cursor ******************************************
*
* NAME
*   Cursor -- (V1)
*
* SYNOPSIS
*   VOID Cursor(struct ACEpansionPlugin *plugin)
*
* FUNCTION
*   Function which is called everytime the cursor signal from the CRTC is
*   active.
*
* INPUTS
*
* RESULT
*
* SEE ALSO
*
*****************************************************************************
*
*/

VOID Cursor(UNUSED struct ACEpansionPlugin *plugin)
{
}



/****** amxmouse.acepansion/AcknowledgeInterrupt ****************************
*
* NAME
*   AcknowledgeInterrupt -- (V1)
*
* SYNOPSIS
*   VOID AcknowledgeInterrupt(struct ACEpansionPlugin *plugin, UBYTE *ivr)
*
* FUNCTION
*   Function which is called everytime the Z80 is entering an interrupt.
*
*   This happens just before the Z80 is actually entering interrupt routine, and
*   it is the good place to provide support for any expansion supporting
*   interrupt vertorized mode (IM 2).
*
* INPUTS
*   *ivr
*     IVR could be set to the desired value (Interrupt Vector Register).
*
* RESULT
*   ReturnInterrupt()
*
* SEE ALSO
*
*****************************************************************************
*
*/

VOID AcknowledgeInterrupt(UNUSED struct ACEpansionPlugin *plugin, UNUSED UBYTE *ivr)
{
}



/****** amxmouse.acepansion/Joystick ****************************************
*
* NAME
*   Joystick -- (V2)
*
* SYNOPSIS
*   VOID Joystick(struct ACEpansionPlugin *plugin, BOOL com1, BOOL com2, UBYTE *ioData)
*
* FUNCTION
*   Function which is called everytime the CPC is accessing the joystick port.
*   Please note that the joystick port in R/W but you cannot know if the CPC
*   is actually reading or writing to it.
*
* INPUTS
*   com1
*     If com1 is active then the CPC is requesting access to joystick 0.
*   com2
*     If com2 is active then the CPC is requesting access to joystick 1.
*   *ioData
*     If the CPC is reading the joystick port (regular case) ioData will be set
*     to 0xff and you are supposed to set to 0 the bits corresponding to the
*     requested joystick (you should use the ACE_IODATAF_JOYSTICK_XXX defines).
*     Please note the additional ACE_IODATAF_JOYSTICK_PAUSE which is not
*     existing on CPC but that you could use to map the play/pause button of
*     a joystick used in ACE on the 'P' key of the keyboard (or the GX4000
*     pause button).
*     If the CPC is writing the joystick port the ioData will contain the
*     written value to the port and modifying it will have not effect (on a
*     real CPC it could simply destroy the PSG chipset!).
*
* RESULT
*
* SEE ALSO
*
*****************************************************************************
*
*/

VOID Joystick(UNUSED struct ACEpansionPlugin *plugin, BOOL com1, BOOL com2, UBYTE *ioData)
{
    struct PluginData *myPlugin = (struct PluginData *)plugin;

#ifdef __HAIKU__
    if (ToggleMouse()) {
        myPlugin->mouseActive = !myPlugin->mouseActive;
    }

    if (myPlugin->mouseActive && com1) {
        GetMouseState(&myPlugin->mouseDeltaX, &myPlugin->mouseDeltaY,
            &myPlugin->mouseRight, &myPlugin->mouseLeft);
    }
#endif

    if(com1)
    {
        if(myPlugin->mouseRight)
            *ioData &= ~ACE_IODATAF_JOYSTICK_FIRE1;
        if(myPlugin->mouseLeft)
            *ioData &= ~ACE_IODATAF_JOYSTICK_FIRE2;
        if(myPlugin->mouseMiddle)
            *ioData &= ~ACE_IODATAF_JOYSTICK_FIRE3;

        if(myPlugin->mouseDeltaX > 0)
            *ioData &= ~ACE_IODATAF_JOYSTICK_RIGHT;
        else if(myPlugin->mouseDeltaX < 0)
            *ioData &= ~ACE_IODATAF_JOYSTICK_LEFT;

        if(myPlugin->mouseDeltaY > 0)
            *ioData &= ~ACE_IODATAF_JOYSTICK_UP;
        else if(myPlugin->mouseDeltaY < 0)
            *ioData &= ~ACE_IODATAF_JOYSTICK_DOWN;
    }
    else if(com2)
    {
        myPlugin->mouseDeltaX /= 2;
        myPlugin->mouseDeltaY /= 4;
    }
}




/*
** Plugin specific private functions
*/

#ifndef __HAIKU__
static IPTR inputhandler(VOID)
{
    struct InputEvent *inputEventList = (struct InputEvent *)REG_A0;
    struct PluginData *myPlugin = (struct PluginData *)REG_A1;
    struct InputEvent *inputEvent = inputEventList;
    struct InputEvent *inputEventPrev = NULL;
    BOOL removeEvent;

    while(inputEvent)
    {
        removeEvent = FALSE;

        switch(inputEvent->ie_Class)
        {
        case IECLASS_RAWKEY:
            if(inputEvent->ie_Code == (RAWKEY_RALT|IECODE_UP_PREFIX))
                myPlugin->mouseActive = !myPlugin->mouseActive;
            break;
        case IECLASS_RAWMOUSE:
            if(myPlugin->mouseActive)
            {
                if(inputEvent->ie_Qualifier|IEQUALIFIER_RELATIVEMOUSE)
                {
                    myPlugin->mouseDeltaX += inputEvent->ie_X;
                    myPlugin->mouseDeltaY -= inputEvent->ie_Y;
                }
                switch(inputEvent->ie_Code)
                {
                case IECODE_LBUTTON:
                    myPlugin->mouseLeft = TRUE;
                    break;
                case IECODE_LBUTTON|IECODE_UP_PREFIX:
                    myPlugin->mouseLeft = FALSE;
                    break;
                case IECODE_RBUTTON:
                    myPlugin->mouseRight = TRUE;
                    break;
                case IECODE_RBUTTON|IECODE_UP_PREFIX:
                    myPlugin->mouseRight = FALSE;
                    break;
                case IECODE_MBUTTON:
                    myPlugin->mouseMiddle = TRUE;
                    break;
                case IECODE_MBUTTON|IECODE_UP_PREFIX:
                    myPlugin->mouseMiddle = FALSE;
                    break;
                }
                removeEvent = TRUE;
            }
            break;
        }
        if(removeEvent)
        {
            if(inputEventPrev)
                inputEventPrev->ie_NextEvent = inputEvent->ie_NextEvent;
            else
                inputEventList = inputEvent->ie_NextEvent;
        }
        else
        {
            inputEventPrev = inputEvent;
        }
        inputEvent = inputEvent->ie_NextEvent;
    }

    return (IPTR)inputEventList;
}
#endif

