/****** joysticks.acepansion/--background-- *********************************
*
* DESCRIPTION
*  This plugin provides joysticks emulation to ACE CPC Emulator.
*
* HISTORY
*
*****************************************************************************
*
*/

#include <proto/exec.h>
#include <proto/utility.h>
#ifndef __HAIKU__
#include <proto/sensors.h>
#endif
#include <proto/muimaster.h>

#ifndef __HAIKU__
#include <libraries/sensors.h>
#include <libraries/sensors_hid.h>
#endif

#include <clib/alib_protos.h>

#ifdef __HAIKU__
#include <Joystick.h>
struct BJoystick* OpenJoystick(int index);
void CloseJoystick(struct BJoystick* j);
class BView;
struct PluginData;
BView* MakeView(STRPTR*, PluginData*, int, int);
class JView;
void EnableJoystickWindow(JView* view, bool activate);
#endif

#include <acepansion/lib_header.h>
#include <libraries/acepansion_plugin.h>

#include "acepansion.h"

#ifndef __HAIKU__
#include "subtasks.h"
#endif

#define CATCOMP_NUMBERS
#include "generated/locale_strings.h" /* catalog string ids */



#define SIGPORT(port) (1 << (port)->mp_SigBit)



struct JoySensorsData
{
    APTR childSensors;
    APTR fire1Sensor;
    APTR fire2Sensor;
    APTR fire3Sensor;
    APTR pauseSensor;
    APTR xySensor;
    APTR xyzSensor;
};

struct PluginData
{
    // Plugin generic part (do NEVER put anything before this!)
    struct ACEpansionPlugin common;

    // Joysticks specifics
    ULONG portJoystick0;
    ULONG portJoystick1;
#ifdef __HAIKU__
	struct BJoystick* sensorsJoystick0;
	struct BJoystick* sensorsJoystick1;
#else
    APTR rootSensors;
    APTR *joySensors;
    struct JoySensorsData sensorsJoystick0;
    struct JoySensorsData sensorsJoystick1;
    struct SubTask *notifyTask;
#endif
    Object *CYA_Joy0;
    Object *CYA_Joy1;
    STRPTR *CYA_GroupJoystickPort;
    STRPTR prefsJoy0;
    STRPTR prefsJoy1;
    struct SignalSemaphore *sema;
};


#ifndef __HAIKU__
struct Library *SensorsBase;
struct Library *UtilityBase;
struct Library *MUIMasterBase;
#endif



static VOID initJoystickPluginData(struct PluginData *plugin);
static VOID cleanJoystickPluginData(struct PluginData *plugin);

VOID configureJoystick(struct PluginData *plugin, ULONG joy, ULONG idx);

static ULONG strLen(CONST_STRPTR str);
static STRPTR strDup(CONST_STRPTR str);

#ifndef __HAIKU__
static VOID subtask_notify(VOID);

#define METHOD_CONFIGURE_JOY0 0
#define METHOD_CONFIGURE_JOY1 1
#define METHOD_UPDATE_JOYLIST 2

struct Msg_ConfigureJoystick
{
    struct PluginData *plugin;
    ULONG idx;
};

static VOID GateHook_DoMethodJoystick(VOID);

static struct EmulLibEntry EmulHook_DoMethodJoystick = { TRAP_LIBNR, 0, (VOID (*)(VOID))GateHook_DoMethodJoystick };
static struct Hook Hook_ConfigureJoytick0  = { { NULL, NULL }, (HOOKFUNC)&EmulHook_DoMethodJoystick, NULL, (APTR)METHOD_CONFIGURE_JOY0 };
static struct Hook Hook_ConfigureJoytick1  = { { NULL, NULL }, (HOOKFUNC)&EmulHook_DoMethodJoystick, NULL, (APTR)METHOD_CONFIGURE_JOY1 };
static struct Hook Hook_UpdateJoystickList = { { NULL, NULL }, (HOOKFUNC)&EmulHook_DoMethodJoystick, NULL, (APTR)METHOD_UPDATE_JOYLIST };

static STRPTR CYA_Dummy[] = { "...", NULL };



/*
** Private function which is called to initialize commons
** just after the library was loaded into memory and prior
** to any other API call.
**
** This is the good place to open our required libraries.
*/
BOOL InitResources(VOID)
{
    SensorsBase = OpenLibrary("sensors.library", 0L);
    UtilityBase = OpenLibrary("utility.library", 0L);
    MUIMasterBase = OpenLibrary("muimaster.library", 0L);

    return MUIMasterBase && UtilityBase && SensorsBase;
}



/*
** Private function which is called to free commons
** just before the library is expurged from memory
**
** This is the good place to close our required libraries.
*/
VOID FreeResources(VOID)
{
    CloseLibrary(MUIMasterBase);
    MUIMasterBase = NULL;

    CloseLibrary(UtilityBase);
    UtilityBase = NULL;

    CloseLibrary(SensorsBase);
    SensorsBase = NULL;
}
#endif



/****** joysticks.library/CreatePlugin **************************************
*
* NAME
*   CreatePlugin -- (V1)
*
* SYNOPSIS
*   struct ACEpansionPlugin * CreatePlugin(CONST_STRPTR *toolTypes, struct SignalSemaphore *sema)
*
* FUNCTION
*   Create the plugin data structure and configure how it will be handled by
*   ACE. This function is usually called at ACE startup.
*
* INPUTS
*   *toolType
*     A pointer to a NULL terminated array of CONST_STRPTR containing the
*     tooltypes provided with the plugin library.
*   *sema
*     Pointer to a semaphore to use to protect access to custom data from
*     ACEpansionPlugin when accessed from subtasks or within your PrefsWindow.
*
* RESULT
*     A pointer to an allocated and initialized plugin data structure.
*
* SEE ALSO
*   DeletePlugin
*   ActivatePlugin
*   GetPrefsPlugin
*
*****************************************************************************
*
*/

struct ACEpansionPlugin * CreatePlugin(CONST_STRPTR *toolTypes, struct SignalSemaphore *sema)
{
    // We are allocating an extended plugin structure to add our custom stuff
    struct PluginData *myPlugin = (struct PluginData*)AllocVec(sizeof(struct PluginData), MEMF_PUBLIC|MEMF_CLEAR);

    if(myPlugin)
    {
        CONST_STRPTR joy0 = NULL;
        CONST_STRPTR joy1 = NULL;

        // Get tooltypes to configure ourself
        if(toolTypes) while(*toolTypes)
        {
            CONST_STRPTR ptr = *toolTypes;
            CONST_STRPTR name = ptr;
            ULONG len = 0;

            if(*ptr != '(')
            {
                while(*ptr != '=' && *ptr != '\0')
                {
                    ptr++;
                    len++;
                }

                if(*ptr == '=')
                {
                    if(!Strnicmp(name,"JOYSTICK0",len))
                    {
                        CONST_STRPTR value = ptr+1;

                        if(*value != '\0')
                        {
                            joy0 = value;
                        }
                    }
                    else if(!Strnicmp(name,"JOYSTICK1",len))
                    {
                        CONST_STRPTR value = ptr+1;

                        if(*value != '\0')
                        {
                            joy1 = value;
                        }
                    }
                }
            }
            toolTypes++;
        }

        // Initialize plugin specifics
        myPlugin->sema = sema;
        myPlugin->prefsJoy0 = strDup(joy0);
        myPlugin->prefsJoy1 = strDup(joy1);
        initJoystickPluginData(myPlugin);

        // Initialize plugin generic part
        myPlugin->common.ap_APIVersion     = API_VERSION;
        myPlugin->common.ap_Flags          = ACE_FLAGSF_ACTIVE_JOYSTICK; // We need the joysticks API (but we are not exclusive)
        myPlugin->common.ap_Title          = GetString(MSG_TITLE);
        myPlugin->common.ap_HelpFileName   = LIBNAME".guide";
        myPlugin->common.ap_ToggleMenuName = GetString(MSG_MENU_TOGGLE);
        myPlugin->common.ap_PrefsMenuName  = GetString(MSG_MENU_PREFS);
#ifdef __HAIKU__
        myPlugin->common.ap_PrefsWindow    = MakeView(myPlugin->CYA_GroupJoystickPort, myPlugin,
                                                 myPlugin->portJoystick0, myPlugin->portJoystick1);
#else
        myPlugin->common.ap_PrefsWindow    = MUI_NewObject(MUIC_Group,
            MUIA_Group_Columns, 2,
            MUIA_Disabled, TRUE,
            MUIA_ShortHelp, (IPTR)GetString(MSG_CYCLE_BUBBLE_HELP),
            GroupFrameT((IPTR)GetString(MSG_GROUP_TITLE)),
            Child, (IPTR)Label2(GetString(MSG_LABEL_JOY0)),
            Child, (IPTR)(myPlugin->CYA_Joy0 = MUI_NewObject(MUIC_Cycle,
                MUIA_Cycle_Entries, (IPTR)myPlugin->CYA_GroupJoystickPort,
                MUIA_Cycle_Active, myPlugin->portJoystick0,
                MUIA_CycleChain, TRUE,
                TAG_END)),
            Child, (IPTR)Label2((IPTR)GetString(MSG_LABEL_JOY1)),
            Child, (IPTR)(myPlugin->CYA_Joy1 = MUI_NewObject(MUIC_Cycle,
                MUIA_Cycle_Entries, (IPTR)myPlugin->CYA_GroupJoystickPort,
                MUIA_Cycle_Active, myPlugin->portJoystick1,
                MUIA_CycleChain, TRUE,
                TAG_END)),
            TAG_END);

        // Finally start our notify subtask
        myPlugin->notifyTask = SpawnSubTask(LIBNAME, subtask_notify, myPlugin);

        DoMethod(myPlugin->CYA_Joy0, MUIM_Notify, MUIA_Cycle_Active, MUIV_EveryTime,
                 MUIV_Notify_Self, 4, MUIM_CallHook, (IPTR)&Hook_ConfigureJoytick0, (IPTR)myPlugin, MUIV_TriggerValue);
        DoMethod(myPlugin->CYA_Joy1, MUIM_Notify, MUIA_Cycle_Active, MUIV_EveryTime,
                 MUIV_Notify_Self, 4, MUIM_CallHook, (IPTR)&Hook_ConfigureJoytick1, (IPTR)myPlugin, MUIV_TriggerValue);
#endif
    }

    return (struct ACEpansionPlugin *)myPlugin;
}



/****** joysticks.library/DeletePlugin **************************************
*
* NAME
*   DeletePlugin -- (V1)
*
* SYNOPSIS
*   VOID DeletePlugin(struct ACEpansionPlugin *plugin)
*
* FUNCTION
*   Delete a plugin data structure. This function is usually called when ACE
*   is exiting.
*
* INPUTS
*   *plugin
*     Pointer to the plugin data structure to delete.
*
* RESULT
*
* SEE ALSO
*   CreatePlugin
*   ActivatePlugin
*
*****************************************************************************
*
*/

VOID DeletePlugin(struct ACEpansionPlugin *plugin)
{
    struct PluginData *myPlugin = (struct PluginData *)plugin;

#ifndef __HAIKU__
    if(myPlugin->notifyTask)
        KillSubTask(myPlugin->notifyTask);
#endif

    FreeVec(myPlugin->prefsJoy0);
    FreeVec(myPlugin->prefsJoy1);

#ifndef __HAIKU__
    MUI_DisposeObject(myPlugin->common.ap_PrefsWindow);
#endif

    cleanJoystickPluginData(myPlugin);

    FreeVec(myPlugin);
}



/****** joysticks.library/ActivatePlugin ************************************
*
* NAME
*   ActivatePlugin -- (V1)
*
* SYNOPSIS
*   BOOL ActivatePlugin(struct ACEpansionPlugin *plugin, BOOL status)
*
* FUNCTION
*   This function is called everytime ACE wants to activate or disactivate your plugin.
*   While you are disactivated, no other function (except Delete) will the called by ACE.
*   Please note that a plugin is always disactivated by ACE prior to DeletePlugin().
*
* INPUTS
*   activate
*     New activation status requested by ACE.
*
* RESULT
*   Actual activation status (you may fail at activation for some reason, but you
*   should never fail at disactivation request).
*
* SEE ALSO
*   CreatePlugin
*   DeletePlugin
*
*****************************************************************************
*
*/

BOOL ActivatePlugin(UNUSED struct ACEpansionPlugin *plugin, BOOL activate)
{
    struct PluginData *myPlugin = (struct PluginData *)plugin;

#ifdef __HAIKU__
	EnableJoystickWindow((JView*)myPlugin->common.ap_PrefsWindow, activate);
#else
    DoMethod(myPlugin->common.ap_PrefsWindow, MUIM_Set, MUIA_Disabled, !activate);
#endif

    return activate;
}



/****** joysticks.acepansion/GetPrefsPlugin *********************************
*
* NAME
*   GetPrefsPlugin -- (V5)
*
* SYNOPSIS
*   STRPTR * GetPrefsPlugin(struct ACEpansionPlugin *plugin)
*
* FUNCTION
*   Function that is used by ACE to know about the current preferences of the
*   plugins in order to save them.
*
* INPUTS
*   *plugin
*     Pointer to the plugin data structure to use.
*   pool
*     Memory pool to use to allocate the tooltypes.
*
* RESULT
*   Pointer to a NULL terminated string array containing the tooltypes or NULL
*   when no tooltypes.
*
* SEE ALSO
*   CreatePlugin
*
*****************************************************************************
*
*/

STRPTR * GetPrefsPlugin(UNUSED struct ACEpansionPlugin *plugin, UNUSED APTR pool)
{
    return NULL;
}



/****** joysticks.library/Reset *********************************************
*
* NAME
*   Reset -- (V1)
*
* SYNOPSIS
*   VOID Reset(struct ACEpansionPlugin *plugin)
*
* FUNCTION
*   Function that is called everytime ACE is issuing a bus reset.
*
* INPUTS
*   *plugin
*     Pointer to the plugin data structure to use.
*
* RESULT
*
* SEE ALSO
*   Emulate
*
*****************************************************************************
*
*/

VOID Reset(UNUSED struct ACEpansionPlugin *plugin)
{
}



/****** joysticks.library/Emulate *******************************************
*
* NAME
*   Emulate -- (V1)
*
* SYNOPSIS
*   VOID Emulate(struct ACEpansionPlugin *plugin, ULONG *signals)
*
* FUNCTION
*   Function that is called at each microsecond of the emulation step inside ACE.
*   Note: Implement this function only if is it really required for your plugin because it
*   is obviously CPU intensive.
*
* INPUTS
*   *plugin
*     Pointer to the plugin data structure to use.
*   *signals
*     Pointer to the I/O signals from the expansion port you can use or alter depending
*     on what you are emulating.
*     The signals bits can be decoded using the following flags:
*       ACE_SIGNALF_INT
*
* RESULT
*
* SEE ALSO
*   Reset
*
*****************************************************************************
*
*/

VOID Emulate(UNUSED struct ACEpansionPlugin *plugin, UNUSED ULONG *signals)
{
}



/****** joysticks.library/WriteIO *******************************************
*
* NAME
*   WriteIO -- (V1)
*
* SYNOPSIS
*   VOID WriteIO(struct ACEpansionPlugin *plugin, USHORT port, UBYTE value)
*
* FUNCTION
*   Function that is called everytime Z80 is performing a I/O port write operation.
*   Note: Implement this function only if is it really required for your plugin because it
*   is obviously CPU intensive.
*
* INPUTS
*   port
*     Z80 port on with the I/O write operation was issued.
*   value
*     Value that was issued on the port.
*
* RESULT
*
* SEE ALSO
*   ReadIO
*
*****************************************************************************
*
*/

VOID WriteIO(UNUSED struct ACEpansionPlugin *plugin, UNUSED USHORT port, UNUSED UBYTE value)
{
}



/****** joysticks.library/ReadIO ********************************************
*
* NAME
*   ReadIO -- (V1)
*
* SYNOPSIS
*   VOID ReadIO(struct ACEpansionPlugin *plugin, USHORT port, UBYTE *value)
*
* FUNCTION
*   Function that is called everytime Z80 is performing a I/O port read operation.
*   Note: Implement this function only if is it really required for your plugin because it
*   is obviously CPU intensive.
*
* INPUTS
*   port
*     Z80 port on with the I/O read operation was issued.
*   *value
*     Value to be returned.
*
* RESULT
*
* SEE ALSO
*   WriteIO
*
*****************************************************************************
*
*/

VOID ReadIO(UNUSED struct ACEpansionPlugin *plugin, UNUSED USHORT port, UNUSED UBYTE *value)
{
}



/****** joysticks.library/GetAudioSample ************************************
*
* NAME
*   GetAudioSample -- (V1)
*
* SYNOPSIS
*   VOID GetAudio(struct ACEpansionPlugin *plugin, SHORT *leftSample, SHORT *rightSample)
*
* FUNCTION
*   Function to be used when your plugin needs to mix an audio output with the one
*   from ACE.
*   It is called at each PSG emulation step, which means at 125KHz.
*
* INPUTS
*   *leftSample
*     Pointer to store the left audio sample to be mixed with ACE audio.
*   *rightSample
*     Pointer to store the right audio sample to be mixed with ACE audio.
*
* RESULT
*
* SEE ALSO
*
*****************************************************************************
*
*/

VOID GetAudio(UNUSED struct ACEpansionPlugin *plugin, UNUSED SHORT *leftSample, UNUSED SHORT *rightSample)
{
}



/****** joysticks.library/Printer *******************************************
*
* NAME
*   Printer -- (V1)
*
* SYNOPSIS
*   VOID WritePrinter(struct ACEpansionPlugin *plugin, UBYTE data, BOOL strobe, BOOL *busy)
*
* FUNCTION
*   Function that is called everytime the CPC is accessing the printer port.
*
* INPUTS
*   data
*     8 bits written value (note that CPC is limited to 7 bits values, only Amstrad Plus
*     can actually provide the 8th bit).
*   strobe
*     Status of the /STROBE signal from the printer port.
*   *busy
*     Busy signal value that can be altered.
*
* RESULT
*
* SEE ALSO
*
*****************************************************************************
*
*/

VOID Printer(UNUSED struct ACEpansionPlugin *plugin, UNUSED UBYTE data, UNUSED BOOL strobe, UNUSED BOOL *busy)
{
}



/****** joysticks.library/Cursor ********************************************
*
* NAME
*   Cursor -- (V1)
*
* SYNOPSIS
*   VOID Cursor(struct ACEpansionPlugin *plugin)
*
* FUNCTION
*   Function which is called everytime the cursor signal from the CRTC is
*   active.
*
* INPUTS
*
* RESULT
*
* SEE ALSO
*
*****************************************************************************
*
*/

VOID Cursor(UNUSED struct ACEpansionPlugin *plugin)
{
}



/****** joysticks.library/AcknowledgeInterrupt ******************************
*
* NAME
*   AcknowledgeInterrupt -- (V1)
*
* SYNOPSIS
*   VOID AcknowledgeInterrupt(struct ACEpansionPlugin *plugin, UBYTE *ivr)
*
* FUNCTION
*   Function which is called everytime the Z80 is entering an interrupt.
*
* INPUTS
*   *ivr
*     IVR could be set to the desired value (Interrupt Vector Register).
*
* RESULT
*
* SEE ALSO
*   ReturnInterrupt
*
*****************************************************************************
*
*/

VOID AcknowledgeInterrupt(UNUSED struct ACEpansionPlugin *plugin, UNUSED UBYTE *ivr)
{
}



/****** joysticks.library/Joystick ******************************************
*
* NAME
*   Joystick -- (V2)
*
* SYNOPSIS
*   VOID Joystick(struct ACEpansionPlugin *plugin, BOOL com1, BOOL com2, UBYTE *ioData)
*
* FUNCTION
*   Function which is called everytime the CPC is accessing the joystick port.
*   Please note that the joystick port in R/W but you cannot know if the CPC
*   is actually reading or writing to it.
*
* INPUTS
*   com1
*     If com1 is active then the CPC is requesting access to joystick 0.
*   com2
*     If com2 is active then the CPC is requesting access to joystick 1.
*   *ioData
*     If the CPC is reading the joystick port (regular case) ioData will be set
*     to 0xff and you are supposed to set to 0 the bits corresponding to the
*     requested joystick (you should use the ACE_IODATAF_JOYSTICK_XXX defines).
*     Please note the additional ACE_IODATAF_JOYSTICK_PAUSE which is not
*     existing on CPC but that you could use to map the play/pause button of
*     a joystick used in ACE on the 'P' key of the keyboard (or the GX4000
*     pause button).
*     If the CPC is writing the joystick port the ioData will contain the
*     written value to the port and modifying it will have not effect (on a
*     real CPC it could simply destroy the PSG chipset!).
*
* RESULT
*
* SEE ALSO
*
*****************************************************************************
*
*/

#ifndef __HAIKU__
#define SET_JOY_BTN(sensor, flag) do { \
    if(sensor) \
    { \
        DOUBLE value; \
        if(GetSensorAttrTags((sensor), SENSORS_HIDInput_Value, (IPTR)&value, TAG_DONE)) \
        { \
            if(value >= 0.5) { *ioData &= ~(flag); } \
        } \
    } \
} while(0)

#define SET_JOY_XY(sensor) do { \
    if(sensor) \
    { \
        DOUBLE ns, ew; \
        if(GetSensorAttrTags((sensor), \
            SENSORS_HIDInput_NS_Value, (IPTR)&ns, \
            SENSORS_HIDInput_EW_Value, (IPTR)&ew, \
            TAG_DONE)) \
        { \
            if(ns <= -0.5) { *ioData &= ~ACE_IODATAF_JOYSTICK_UP; } \
            if(ns >=  0.5) { *ioData &= ~ACE_IODATAF_JOYSTICK_DOWN; } \
            if(ew <= -0.5) { *ioData &= ~ACE_IODATAF_JOYSTICK_LEFT; } \
            if(ew >=  0.5) { *ioData &= ~ACE_IODATAF_JOYSTICK_RIGHT; } \
        } \
    } \
} while(0)

#define SET_JOY_XYZ(sensor) do { \
    if(sensor) \
    { \
        DOUBLE x, y; \
        if(GetSensorAttrTags((sensor), \
            SENSORS_HIDInput_X_Index, (IPTR)&x, \
            SENSORS_HIDInput_Y_Index, (IPTR)&y, \
            TAG_DONE)) \
        { \
            if(y <= -0.5) { *ioData &= ~ACE_IODATAF_JOYSTICK_UP; } \
            if(y >=  0.5) { *ioData &= ~ACE_IODATAF_JOYSTICK_DOWN; } \
            if(x <= -0.5) { *ioData &= ~ACE_IODATAF_JOYSTICK_LEFT; } \
            if(x >=  0.5) { *ioData &= ~ACE_IODATAF_JOYSTICK_RIGHT; } \
        } \
    } \
} while(0)
#endif

VOID Joystick(struct ACEpansionPlugin *plugin, BOOL com1, BOOL com2, UBYTE *ioData)
{
    struct PluginData *myPlugin = (struct PluginData *)plugin;

    if(com1 || com2)
    {
#ifdef __HAIKU__
		BJoystick* j = com1 ? myPlugin->sensorsJoystick0 : myPlugin->sensorsJoystick1;

		if (j) {
			j->Update();
			uint32 buttons = j->ButtonValues();
			if (buttons & 1) *ioData &= ~ACE_IODATAF_JOYSTICK_FIRE1;
			if (buttons & 2) *ioData &= ~ACE_IODATAF_JOYSTICK_FIRE2;
			if (buttons & 4) *ioData &= ~ACE_IODATAF_JOYSTICK_FIRE3;
			if (buttons & 8) *ioData &= ~ACE_IODATAF_JOYSTICK_PAUSE;
			if (j->horizontal < -16) *ioData &= ~ACE_IODATAF_JOYSTICK_LEFT;
			if (j->horizontal > 16) *ioData &= ~ACE_IODATAF_JOYSTICK_RIGHT;
			if (j->vertical < -16) *ioData &= ~ACE_IODATAF_JOYSTICK_UP;
			if (j->vertical > 16) *ioData &= ~ACE_IODATAF_JOYSTICK_DOWN;
		}
#else
        struct JoySensorsData *joySensors = com1 ? &myPlugin->sensorsJoystick0 : &myPlugin->sensorsJoystick1;

        SET_JOY_BTN(joySensors->fire1Sensor, ACE_IODATAF_JOYSTICK_FIRE1);
        SET_JOY_BTN(joySensors->fire2Sensor, ACE_IODATAF_JOYSTICK_FIRE2);
        SET_JOY_BTN(joySensors->fire3Sensor, ACE_IODATAF_JOYSTICK_FIRE3);
        SET_JOY_BTN(joySensors->pauseSensor, ACE_IODATAF_JOYSTICK_PAUSE);
        SET_JOY_XY(joySensors->xySensor);
        SET_JOY_XYZ(joySensors->xyzSensor);
#endif
    }
}



/****** joysticks.acepansion/ReturnInterrupt ********************************
*
* NAME
*   ReturnInterrupt -- (V3)
*
* SYNOPSIS
*   VOID ReturnInterrupt(struct ACEpansionPlugin *plugin)
*
* FUNCTION
*   Function which is called everytime the Z80 is returning from an interrupt
*   using RETI instruction.
*
* INPUTS
*
* RESULT
*
* SEE ALSO
*   AcknowledgeInterrupt
*
*****************************************************************************
*
*/

VOID ReturnInterrupt(UNUSED struct ACEpansionPlugin *plugin)
{
}



/****** joysticks.acepansion/LightDeviceDiodePulse **************************
*
* NAME
*   LightDeviceDiodePulse -- (V4)
*
* SYNOPSIS
*   VOID LightDeviceDiodePulse(struct ACEpansionPlugin *plugin)
*
* FUNCTION
*   This function is called at each photo-diode pulse, when the beam from
*   the screen monitor is just infront of the light device photo-diode.
*
* INPUTS
*
* RESULT
*
* SEE ALSO
*   LightDeviceButton
*
*****************************************************************************
*
*/

VOID LightDeviceDiodePulse(UNUSED struct ACEpansionPlugin *plugin)
{
}



/****** joysticks.acepansion/LightDeviceButton ******************************
*
* NAME
*   LightDeviceButton -- (V4)
*
* SYNOPSIS
*   VOID LightDeviceButton(struct ACEpansionPlugin *plugin, BOOL pressed)
*
* FUNCTION
*   This function is called everytime the state of the light device button
*   is changing.
*
* INPUTS
*   pressed
*     New state of the light device button.
*
* RESULT
*
* NOTE
*   When a light device plugin is activated, ACE is mapping this button on
*   mouse select button (LMB), so that on-screen click will emulate light
*   device button.
*
* SEE ALSO
*   LightDeviceDiodePulse
*
*****************************************************************************
*
*/

VOID LightDeviceButton(UNUSED struct ACEpansionPlugin *plugin, UNUSED BOOL pressed)
{
}



/*
** Private joysticks functions
*/

VOID configureJoystick(struct PluginData *myPlugin, ULONG joy, ULONG idx)
{
#ifdef __HAIKU__
	if (joy == 0) {
		CloseJoystick(myPlugin->sensorsJoystick0);
    	myPlugin->sensorsJoystick0 = OpenJoystick(idx);
		myPlugin->portJoystick0 = idx;
	} else {
		CloseJoystick(myPlugin->sensorsJoystick1);
    	myPlugin->sensorsJoystick1 = OpenJoystick(idx);
		myPlugin->portJoystick1 = idx;
	}
#else
    APTR parentSensor;
    struct JoySensorsData *joySensors;

    if(joy == 0)
    {
        myPlugin->portJoystick0 = idx;
        joySensors = &myPlugin->sensorsJoystick0;
    }
    else
    {
        myPlugin->portJoystick1 = idx;
        joySensors = &myPlugin->sensorsJoystick1;
    }

    joySensors->fire1Sensor = NULL;
    joySensors->fire2Sensor = NULL;
    joySensors->fire3Sensor = NULL;
    joySensors->pauseSensor = NULL;
    joySensors->xySensor = NULL;
    joySensors->xyzSensor = NULL;

    if(joySensors->childSensors)
    {
        ReleaseSensorsList(joySensors->childSensors, NULL);
        joySensors->childSensors = NULL;
    }

    parentSensor = myPlugin->joySensors[idx];

    if(parentSensor)
    {
        joySensors->childSensors = ObtainSensorsListTags(
            SENSORS_Parent, (IPTR)parentSensor,
            SENSORS_Class, SensorClass_HID,
            TAG_DONE);

        if(joySensors->childSensors)
        {
            APTR sensor = NULL;
            ULONG triggerCount = 0;
            ULONG analogCount = 0;

            while((sensor = NextSensor(sensor, joySensors->childSensors, NULL)) != NULL)
            {
                LONG type = -1;
                LONG limb = -1;

                if(GetSensorAttrTags(sensor,
                    SENSORS_Type, (IPTR)&type,
                    SENSORS_HIDInput_Limb, (IPTR)&limb,
                    TAG_DONE))
                {
                    switch(type)
                    {
                        case SensorType_HIDInput_Trigger:
                            switch(triggerCount)
                            {
                                // Allocate fire 2 first because it is actually the main CPC fire
                                case 0: joySensors->fire2Sensor = sensor; triggerCount++; break;
                                case 1: joySensors->fire1Sensor = sensor; triggerCount++; break;
                                case 2: joySensors->fire3Sensor = sensor; triggerCount++; break;
                                case 3: joySensors->pauseSensor = sensor; triggerCount++; break;
                            }
                            break;
                        case SensorType_HIDInput_Analog:
                            switch(analogCount)
                            {
                                // Use analog inputs only if trigger were not found
                                case 0: if(joySensors->fire2Sensor == NULL) { joySensors->fire2Sensor = sensor; analogCount++; } break;
                                case 1: if(joySensors->fire1Sensor == NULL) { joySensors->fire1Sensor = sensor; analogCount++; } break;
                                case 2: if(joySensors->fire3Sensor == NULL) { joySensors->fire3Sensor = sensor; analogCount++; } break;
                                case 3: if(joySensors->pauseSensor == NULL) { joySensors->pauseSensor = sensor; analogCount++; } break;
                            }
                            break;
                        case SensorType_HIDInput_Stick:
                        case SensorType_HIDInput_AnalogStick:
                            switch(limb)
                            {
                                // Give priority to the left xy sensor in case of several
                                case Sensor_HIDInput_Limb_LeftHand: joySensors->xySensor = sensor; break;
                                default: if(joySensors->xySensor == NULL) { joySensors->xySensor = sensor; } break;
                            }
                            break;
                        case SensorType_HIDInput_3DStick:
                            // We should have only one 3D stick
                            joySensors->xyzSensor = sensor;
                            break;
                    }
                }
            }
        }
    }
#endif
}

static VOID initJoystickPluginData(struct PluginData *myPlugin)
{
#ifdef __HAIKU__
	BJoystick j;
	ULONG count = j.CountDevices();
#else
    struct TagItem sensorTags[] =
    {
        { SENSORS_Class, SensorClass_HID },
        // { SENSORS_Type, SensorType_HID_Gamepad },
        { TAG_END }
    };
    ULONG count = GetSensorsCount(sensorTags);
#endif
    ULONG joy0 = 0;
    ULONG joy1 = 0;

    myPlugin->CYA_GroupJoystickPort = (STRPTR*)AllocVec((count+2) * sizeof(STRPTR), MEMF_ANY|MEMF_CLEAR);
#ifndef __HAIKU__
    myPlugin->joySensors = AllocVec((count+1) * sizeof(APTR), MEMF_ANY|MEMF_CLEAR);
    myPlugin->rootSensors = ObtainSensorsList(sensorTags);
    myPlugin->sensorsJoystick0.childSensors = NULL;
    myPlugin->sensorsJoystick1.childSensors = NULL;
#endif
    myPlugin->portJoystick0 = 0;
    myPlugin->portJoystick1 = 0;

#ifdef __HAIKU__
    if(myPlugin->CYA_GroupJoystickPort)
	{
		myPlugin->CYA_GroupJoystickPort[0] = strDup(GetString(MSG_NO_JOYSTICK));

		for (int i = 0; i < count; i++)
		{
			char sensorName[B_OS_NAME_LENGTH];
			status_t status = j.GetDeviceName(i, sensorName);
			myPlugin->CYA_GroupJoystickPort[i + 1] = ((status == B_OK) && sensorName[0] != '\0') ? strDup(sensorName) : strDup(GetString(MSG_UNNAMED_JOYSTICK));

			if (status == B_OK) {
				if((joy0 == 0 && i == 0) // Default to 1st found joytick
						|| (myPlugin->prefsJoy0 && Stricmp(myPlugin->prefsJoy0, sensorName) == 0))
				{
					joy0 = i;
				}
				if((joy1 == 0 && i == 1) // Default to 2nd found joytick
						|| (myPlugin->prefsJoy1 && Stricmp(myPlugin->prefsJoy1, sensorName) == 0))
				{
					joy1 = i;
				}
			}
		}
	}
#else
    if(myPlugin->CYA_GroupJoystickPort
    && myPlugin->joySensors)
    {
        myPlugin->CYA_GroupJoystickPort[0] = strDup(GetString(MSG_NO_JOYSTICK));

        if(myPlugin->rootSensors)
        {
            ULONG idx = 1;
            APTR sensor = NULL;

            while(idx < count + 1
              && (sensor = NextSensor(sensor, myPlugin->rootSensors, NULL)) != NULL)
            {
                STRPTR sensorName = NULL;

                GetSensorAttrTags(sensor,
                    SENSORS_HID_Name, (IPTR)&sensorName,
                    TAG_DONE);

                myPlugin->CYA_GroupJoystickPort[idx] = (sensorName && sensorName[0] != '\0') ? strDup(sensorName) : strDup(GetString(MSG_UNNAMED_JOYSTICK));
                myPlugin->joySensors[idx] = sensor;

                // Check if the sensor is matching preferences
                if(sensorName)
                {
                    if((joy0 == 0 && idx == 1) // Default to 1st found joytick
                    || (myPlugin->prefsJoy0 && Stricmp(myPlugin->prefsJoy0, sensorName) == 0))
                    {
                        joy0 = idx;
                    }
                    if((joy1 == 0 && idx == 2) // Default to 2nd found joytick
                    || (myPlugin->prefsJoy1 && Stricmp(myPlugin->prefsJoy1, sensorName) == 0))
                    {
                        joy1 = idx;
                    }
                }
                idx++;
            }
        }
    }
    else
    {
        // Something went wrong...
        FreeVec(myPlugin->CYA_GroupJoystickPort);
        myPlugin->CYA_GroupJoystickPort = NULL;
        FreeVec(myPlugin->joySensors);
        myPlugin->joySensors = NULL;
    }
#endif

    configureJoystick(myPlugin, 0, joy0);
    configureJoystick(myPlugin, 1, joy1);
}

static VOID cleanJoystickPluginData(struct PluginData *myPlugin)
{
    STRPTR *cursor = myPlugin->CYA_GroupJoystickPort;

    if(cursor) while(*cursor)
    {
        FreeVec(*cursor);
        cursor++;
    }

#ifndef __HAIKU__
    if(myPlugin->sensorsJoystick0.childSensors)
        ReleaseSensorsList(myPlugin->sensorsJoystick0.childSensors, NULL);

    if(myPlugin->sensorsJoystick1.childSensors)
        ReleaseSensorsList(myPlugin->sensorsJoystick1.childSensors, NULL);

    if(myPlugin->rootSensors)
        ReleaseSensorsList(myPlugin->rootSensors, NULL);
#endif

    FreeVec(myPlugin->CYA_GroupJoystickPort);

#ifndef __HAIKU__
    FreeVec(myPlugin->joySensors);
#endif
}

#ifndef __HAIKU__
static VOID GateHook_DoMethodJoystick(VOID)
{
    struct Hook *hook = (struct Hook *)REG_A0;
    struct Msg_ConfigureJoystick *msg = (struct Msg_ConfigureJoystick *)REG_A1;

    ObtainSemaphore(msg->plugin->sema);

    switch((ULONG)hook->h_Data)
    {
        case METHOD_CONFIGURE_JOY0:
            configureJoystick(msg->plugin, 0, msg->idx);
            break;
        case METHOD_CONFIGURE_JOY1:
            configureJoystick(msg->plugin, 1, msg->idx);
            break;
        case METHOD_UPDATE_JOYLIST:
            DoMethod(msg->plugin->CYA_Joy0, MUIM_NoNotifySet, MUIA_Cycle_Active, 0);
            DoMethod(msg->plugin->CYA_Joy1, MUIM_NoNotifySet, MUIA_Cycle_Active, 0);
            DoMethod(msg->plugin->CYA_Joy0, MUIM_NoNotifySet, MUIA_Cycle_Entries, (IPTR)CYA_Dummy);
            DoMethod(msg->plugin->CYA_Joy1, MUIM_NoNotifySet, MUIA_Cycle_Entries, (IPTR)CYA_Dummy);
            cleanJoystickPluginData(msg->plugin);
            initJoystickPluginData(msg->plugin);
            DoMethod(msg->plugin->CYA_Joy0, MUIM_NoNotifySet, MUIA_Cycle_Entries, (IPTR)msg->plugin->CYA_GroupJoystickPort);
            DoMethod(msg->plugin->CYA_Joy1, MUIM_NoNotifySet, MUIA_Cycle_Entries, (IPTR)msg->plugin->CYA_GroupJoystickPort);
            DoMethod(msg->plugin->CYA_Joy0, MUIM_NoNotifySet, MUIA_Cycle_Active, msg->plugin->portJoystick0);
            DoMethod(msg->plugin->CYA_Joy1, MUIM_NoNotifySet, MUIA_Cycle_Active, msg->plugin->portJoystick1);
            break;
    }

    ReleaseSemaphore(msg->plugin->sema);
}

static VOID subtask_notify(VOID)
{
    struct SubTask *subTask;

    if((subTask = InitSubTask()))
    {
        BOOL isRunning = TRUE;
        struct PluginData *plugin = (struct PluginData *)subTask->st_Data;
        struct Library *SensorsBase = OpenLibrary("sensors.library", 0L);
        struct MsgPort *notifyPort = CreateMsgPort();
        struct Message *notifyMsg;
        APTR sensorsNotify = NULL;
        struct SubTaskMsg *taskMsg;

        if(SensorsBase && notifyPort)
        {
            sensorsNotify = StartSensorNotifyTags(NULL,
                SENSORS_Class, SensorClass_HID,
                SENSORS_Notification_ClassListChanged, TRUE,
                SENSORS_Notification_Destination, (IPTR)notifyPort);
        }

        // After the sub task is up and running, we go into
        // a loop and process the messages from the main task.
        for(;;)
        {
            // Manage the messages received by the subtask...
            while((taskMsg = (struct SubTaskMsg *)GetMsg(subTask->st_Port)))
            {
                switch(taskMsg->stm_Command)
                {
                    case STC_SHUTDOWN:
                        // This is the shutdown message from KillSubTask().
                        isRunning = FALSE;
                        break;
                }

                // If we received a shutdown message, we do not reply it immediately.
                // First, we need to free our resources.
                if(!isRunning) break;

                ReplyMsg((struct Message *)taskMsg);
            }
            // It's over...
            if(!isRunning) break;

            if(notifyPort) while((notifyMsg = GetMsg(notifyPort)))
            {
                // Whatever the message is, we know we just have to update the joysticks list.
                if(plugin->common.ap_PrefsWindow)
                {
                    DoMethod(_app(plugin->common.ap_PrefsWindow), MUIM_Application_PushMethod,
                             (IPTR)plugin->common.ap_PrefsWindow, 3, MUIM_CallHook, (IPTR)&Hook_UpdateJoystickList, (IPTR)plugin);
                }
                ReplyMsg(notifyMsg);
            }

            // Let's go!
            Wait(SIGPORT(subTask->st_Port) | SIGPORT(notifyPort));
        }

        if(sensorsNotify)
            EndSensorNotify(sensorsNotify, NULL);

        if(notifyPort)
        {
            while((notifyMsg = GetMsg(notifyPort))) ReplyMsg(notifyMsg);
            DeleteMsgPort(notifyPort);
        }

        CloseLibrary(SensorsBase);

        ExitSubTask(subTask, taskMsg);
    }
}
#endif



/*
** Private helper functions
*/

static ULONG strLen(CONST_STRPTR str)
{
    ULONG count = 0;

    while(*str++ != '\0') count++;

    return count;
}

static STRPTR strDup(CONST_STRPTR str)
{
    STRPTR newStr = NULL;

    if(str)
    {
        ULONG size = strLen(str)+1;
        if((newStr = (STRPTR)AllocVec(size, MEMF_ANY)))
        {
            CopyMem((CONST APTR)str, newStr, size);
        }
    }
    return newStr;
}

