/****** printer.acepansion/--background-- **********************************
*
* DESCRIPTION
*  This plugin provides printer emulation to ACE CPC Emulator, through a file
*  redirection.
*
* HISTORY
*
*****************************************************************************
*
*/

#include <proto/exec.h>
#include <proto/utility.h>
#include <proto/dos.h>
#include <proto/intuition.h>
#include <proto/muimaster.h>

#include <libraries/asl.h>
#include <clib/alib_protos.h>

#include <acepansion/lib_header.h>
#include <libraries/acepansion_plugin.h>

#include "acepansion.h"
#define CATCOMP_NUMBERS
#include "generated/locale_strings.h" /* catalog string ids */

#ifdef __HAIKU__
#include <Beep.h>
#include <Box.h>

#include "PopASL.h"
#endif


struct PluginData
{
    // Plugin generic part (do NEVER put anything before this!)
    struct ACEpansionPlugin common;

    // Printer specifics
    struct SignalSemaphore *sema;
    BOOL   isActivated;
    STRPTR printerFileName;
    BPTR   printerFile;
    BOOL   previousStrobe;
};



struct Library *DOSBase;
struct Library *UtilityBase;
struct Library *IntuitionBase;
struct Library *MUIMasterBase;



static ULONG strLen(CONST_STRPTR str);
static STRPTR strDup(CONST_STRPTR str);

struct Msg_SetPrinterFileName
{
    struct PluginData *plugin;
    STRPTR fileName;
};

static BOOL smartOpenPrinterFile(struct PluginData *plugin);

#ifdef __HAIKU__
static VOID gateHook_SetPrinterFileName(struct PluginData* plugin, const char* fileName);
#else
static VOID gateHook_SetPrinterFileName(VOID);
#endif

#ifndef __HAIKU__
static struct EmulLibEntry emulHook_SetPrinterFileName = { TRAP_LIBNR, 0, (VOID (*)(VOID))gateHook_SetPrinterFileName };
static struct Hook hook_SetPrinterFileName  = { { NULL, NULL }, (HOOKFUNC)&emulHook_SetPrinterFileName, NULL, NULL };
#endif



/*
** Private function which is called to initialize commons
** just after the library was loaded into memory and prior
** to any other API call.
**
** This is the good place to open our required libraries.
*/
BOOL InitResources(VOID)
{
#ifdef __HAIKU__
    return 1;
#else
    DOSBase = OpenLibrary("dos.library", 0L);
    UtilityBase = OpenLibrary("utility.library", 0L);
    IntuitionBase = OpenLibrary("intuition.library", 0L);
    MUIMasterBase = OpenLibrary("muimaster.library", 0L);

    return (DOSBase && UtilityBase && IntuitionBase && MUIMasterBase);
#endif
}



/*
** Private function which is called to free commons
** just before the library is expurged from memory
**
** This is the good place to close our required libraries.
*/
VOID FreeResources(VOID)
{
#ifndef __HAIKU__
    CloseLibrary(DOSBase);
    CloseLibrary(UtilityBase);
    CloseLibrary(IntuitionBase);
    CloseLibrary(MUIMasterBase);
#endif
}



/****** printer.acepansion/CreatePlugin *************************************
*
* NAME
*   CreatePlugin -- (V1)
*
* SYNOPSIS
*   struct ACEpansionPlugin * CreatePlugin(CONST_STRPTR *toolTypes, struct SignalSemaphore *sema)
*
* FUNCTION
*   Create the plugin data structure and configure how it will be handled by
*   ACE. This function is usually called at ACE startup.
*
* INPUTS
*   *toolType
*     A pointer to a NULL terminated array of CONST_STRPTR containing the
*     tooltypes provided with the plugin library.
*     The ACTIVE tooltype is handled internally by ACE for all plugins, but
*     you can here handle additional tooltypes to configure your plugin.
*   *sema
*     Pointer to a semaphore to use to protect access to custom data from
*     ACEpansionPlugin when accessed from subtasks or within your PrefsWindow.
*
* RESULT
*   A pointer to an allocated and initialized plugin data structure.
*
* SEE ALSO
*   DeletePlugin
*   ActivatePlugin
*   GetPrefsPlugin
*
*****************************************************************************
*
*/

struct ACEpansionPlugin * CreatePlugin(CONST_STRPTR *toolTypes, struct SignalSemaphore *sema)
{
    // We are allocating an extended plugin structure to add our custom stuff
    struct PluginData *myPlugin = (struct PluginData*)AllocVec(sizeof(struct PluginData), MEMF_PUBLIC|MEMF_CLEAR);

    if(myPlugin)
    {
#ifdef __HAIKU__
        CONST_STRPTR fileName = "/tmp/ACE_printer.log";
#else
        CONST_STRPTR fileName = "T:ACE_printer.log";
#endif
        Object *TXT_PrinterFile;

        if(toolTypes) while(*toolTypes)
        {
            CONST_STRPTR ptr = *toolTypes;
            CONST_STRPTR name = ptr;
            ULONG len = 0;

            if(*ptr != '(')
            {
                while(*ptr != '=' && *ptr != '\0')
                {
                    ptr++;
                    len++;
                }

                if(*ptr == '=')
                {
                    if(!Strnicmp(name,"PRINTERFILE",len))
                    {
                        CONST_STRPTR value = ptr+1;

                        if(*value != '\0')
                        {
                            fileName = value;
                        }
                    }
                }
            }
            toolTypes++;
        }

        // Initialize plugin specifics
        myPlugin->sema = sema;
        myPlugin->isActivated = FALSE;
        myPlugin->printerFileName = strDup(fileName);
        myPlugin->printerFile = (BPTR)0;
        myPlugin->previousStrobe = TRUE;

        // Initialize plugin generic part
        myPlugin->common.ap_APIVersion     = API_VERSION;
        myPlugin->common.ap_Flags          = ACE_FLAGSF_EXCLUSIVE_PRINTER  // We must be the only active plugin to use printer port at once
                                           | ACE_FLAGSF_ACTIVE_PRINTER     // We need the printer API
                                           | ACE_FLAGSF_SAVE_PREFS;        // We have preferences to save
        myPlugin->common.ap_Title          = GetString(MSG_TITLE);
        myPlugin->common.ap_HelpFileName   = LIBNAME".guide";
        myPlugin->common.ap_ToggleMenuName = GetString(MSG_MENU_TOGGLE);
        myPlugin->common.ap_PrefsMenuName  = GetString(MSG_MENU_PREFS);
#ifdef __HAIKU__
		BBox* container = new BBox("printer settings");
		container->SetLabel(GetString(MSG_GROUP_TITLE));
        myPlugin->common.ap_PrefsWindow    = container;
        PopASL* asl = new PopASL(0, GetString(MSG_LABEL_FILE), myPlugin->printerFileName,
			NULL, NULL, GetString(MSG_FILE_HELP), B_SAVE_PANEL, true);
		container->AddChild(asl);
        asl->Button()->SetEnabled(false);
        asl->SetCallback((PopASL::HookFunction)&gateHook_SetPrinterFileName, myPlugin);
#else
        myPlugin->common.ap_PrefsWindow    = MUI_NewObject(MUIC_Group,
                MUIA_Group_Columns, 2,
                MUIA_Disabled, TRUE,
                GroupFrameT((IPTR)GetString(MSG_GROUP_TITLE)),
                Child, (IPTR)Label2(GetString(MSG_LABEL_FILE)),
                Child, (IPTR)MUI_NewObject(MUIC_Popasl,
                    MUIA_Popstring_String, (IPTR)(TXT_PrinterFile = MUI_NewObject(MUIC_String,
                        StringFrame,
                        MUIA_String_Contents, (IPTR)myPlugin->printerFileName,
                        MUIA_CycleChain, TRUE,
                        TAG_END)),
                    MUIA_Popstring_Button, (IPTR)MUI_NewObject(MUIC_Text,
                        ButtonFrame,
                        MUIA_Weight, 0,
                        MUIA_InputMode , MUIV_InputMode_RelVerify,
                        MUIA_Background, MUII_ButtonBack,
                        MUIA_CycleChain, TRUE,
                        MUIA_Text_Contents, (IPTR)"\33I[6:19]", // MUII_PopFile
                        TAG_END),
                    ASLFR_TitleText, (IPTR)GetString(MSG_ASL_REQUEST),
                    ASLFR_DoSaveMode, TRUE,
                    MUIA_ShortHelp, (IPTR)GetString(MSG_FILE_HELP),
                    TAG_END),
                TAG_END);

            DoMethod(TXT_PrinterFile, MUIM_Notify, MUIA_String_Acknowledge, MUIV_EveryTime,
               (IPTR)TXT_PrinterFile, 4, MUIM_CallHook, (IPTR)&hook_SetPrinterFileName, (IPTR)myPlugin, MUIV_TriggerValue);
#endif
    }

    return (struct ACEpansionPlugin *)myPlugin;
}



/****** printer.acepansion/DeletePlugin *************************************
*
* NAME
*   DeletePlugin -- (V1)
*
* SYNOPSIS
*   VOID DeletePlugin(struct ACEpansionPlugin *plugin)
*
* FUNCTION
*   Delete a plugin data structure. This function is usually called when ACE
*   is exiting.
*
* INPUTS
*   *plugin
*     Pointer to the plugin data structure to delete.
*
* RESULT
*
* SEE ALSO
*   CreatePlugin
*   ActivatePlugin
*
*****************************************************************************
*
*/

VOID DeletePlugin(struct ACEpansionPlugin *plugin)
{
    struct PluginData *myPlugin = (struct PluginData *)plugin;

#ifdef __HAIKU__
	BView* view = (BView*)myPlugin->common.ap_PrefsWindow;
	view->LockLooper();
	view->RemoveSelf();
	delete view;
#else
    MUI_DisposeObject(myPlugin->common.ap_PrefsWindow);
#endif
    FreeVec(myPlugin->printerFileName);
    FreeVec(myPlugin);
}



/****** printer.acepansion/ActivatePlugin ***********************************
*
* NAME
*   ActivatePlugin -- (V1)
*
* SYNOPSIS
*   BOOL ActivatePlugin(struct ACEpansionPlugin *plugin, BOOL status)
*
* FUNCTION
*   This function is called everytime ACE wants to activate or disactivate your plugin.
*   While you are disactivated, no other function (except Delete) will the called by ACE.
*   Please note that a plugin is always disactivated by ACE prior to DeletePlugin().
*
* INPUTS
*   activate
*     New activation status requested by ACE.
*
* RESULT
*   Actual activation status (you may fail at activation for some reason, but you
*   should never fail at disactivation request).
*
* SEE ALSO
*   CreatePlugin
*   DeletePlugin
*
*****************************************************************************
*
*/

BOOL ActivatePlugin(struct ACEpansionPlugin *plugin, BOOL activate)
{
    struct PluginData *myPlugin = (struct PluginData *)plugin;

	if(activate)
    {
        if(!smartOpenPrinterFile(myPlugin))
        {
            activate = FALSE;
        }
	}
    else
    {
        if(myPlugin->printerFile)
        {
            Close(myPlugin->printerFile);
            myPlugin->printerFile = (BPTR)0;
        }
    }

    myPlugin->isActivated = activate;
#ifdef __HAIKU__
    BView* view = (BView*)myPlugin->common.ap_PrefsWindow;
    if (view->LockLooper() == true) {
        ((PopASL*)view->ChildAt(0))->Button()->SetEnabled(activate);
        view->UnlockLooper();
    }
#else
    DoMethod(myPlugin->common.ap_PrefsWindow, MUIM_Set, MUIA_Disabled, !activate);
#endif

    return activate;
}



/****** printer.acepansion/GetPrefsPlugin ***********************************
*
* NAME
*   GetPrefsPlugin -- (V5)
*
* SYNOPSIS
*   STRPTR * GetPrefsPlugin(struct ACEpansionPlugin *plugin)
*
* FUNCTION
*   Function that is used by ACE to know about the current preferences of the
*   plugins in order to save them.
*
* INPUTS
*   *plugin
*     Pointer to the plugin data structure to use.
*   pool
*     Memory pool to use to allocate the tooltypes.
*
* RESULT
*   Pointer to a NULL terminated string array containing the tooltypes or NULL
*   when no tooltypes.
*
* SEE ALSO
*   CreatePlugin
*
*****************************************************************************
*
*/

STRPTR * GetPrefsPlugin(struct ACEpansionPlugin *plugin, APTR pool)
{
    struct PluginData *myPlugin = (struct PluginData *)plugin;
    STRPTR *tooltypes = (STRPTR*)AllocPooled(pool, 2*sizeof(STRPTR));

    if(tooltypes)
    {
        tooltypes[0] = MakeToolType(pool, "PRINTERFILE", myPlugin->printerFileName);
        tooltypes[1] = NULL;
    }

    return tooltypes;
}



/****** printer.acepansion/Reset ********************************************
*
* NAME
*   Reset -- (V1)
*
* SYNOPSIS
*   VOID Reset(struct ACEpansionPlugin *plugin)
*
* FUNCTION
*   Function that is called everytime ACE is issuing a bus reset.
*
* INPUTS
*   *plugin
*     Pointer to the plugin data structure to use.
*
* RESULT
*
* SEE ALSO
*   Emulate
*
*****************************************************************************
*
*/

VOID Reset(UNUSED struct ACEpansionPlugin *plugin)
{
}



/****** printer.acepansion/Emulate ******************************************
*
* NAME
*   Emulate -- (V1)
*
* SYNOPSIS
*   VOID Emulate(struct ACEpansionPlugin *plugin, ULONG *signals)
*
* FUNCTION
*   Function that is called at each microsecond of the emulation step inside ACE.
*   Note: Implement this function only if is it really required for your plugin because it
*   is obviously CPU intensive.
*
* INPUTS
*   *plugin
*     Pointer to the plugin data structure to use.
*   *signals
*     Pointer to the I/O signals from the expansion port you can use or alter depending
*     on what you are emulating.
*     The signals bits can be decoded using the following flags:
*       ACE_SIGNALF_INT
*
* RESULT
*
* SEE ALSO
*   Reset
*
*****************************************************************************
*
*/

VOID Emulate(UNUSED struct ACEpansionPlugin *plugin, UNUSED ULONG *signals)
{
}



/****** printer.acepansion/WriteIO ******************************************
*
* NAME
*   WriteIO -- (V1)
*
* SYNOPSIS
*   VOID WriteIO(struct ACEpansionPlugin *plugin, USHORT port, UBYTE value)
*
* FUNCTION
*   Function that is called everytime Z80 is performing a I/O port write operation.
*   Note: Implement this function only if is it really required for your plugin because it
*   is obviously CPU intensive.
*
* INPUTS
*   port
*     Z80 port on with the I/O write operation was issued.
*   value
*     Value that was issued on the port.
*
* RESULT
*
* SEE ALSO
*   ReadIO
*
*****************************************************************************
*
*/

VOID WriteIO(UNUSED struct ACEpansionPlugin *plugin, UNUSED USHORT port, UNUSED UBYTE value)
{
}



/****** printer.acepansion/ReadIO *******************************************
*
* NAME
*   ReadIO -- (V1)
*
* SYNOPSIS
*   VOID ReadIO(struct ACEpansionPlugin *plugin, USHORT port, UBYTE *value)
*
* FUNCTION
*   Function that is called everytime Z80 is performing a I/O port read operation.
*   Note: Implement this function only if is it really required for your plugin because it
*   is obviously CPU intensive.
*
* INPUTS
*   port
*     Z80 port on with the I/O read operation was issued.
*   *value
*     Value to be returned.
*
* RESULT
*
* SEE ALSO
*   WriteIO
*
*****************************************************************************
*
*/

VOID ReadIO(UNUSED struct ACEpansionPlugin *plugin, UNUSED USHORT port, UNUSED UBYTE *value)
{
}



/****** printer.acepansion/GetAudioSample ***********************************
*
* NAME
*   GetAudioSample -- (V1)
*
* SYNOPSIS
*   VOID GetAudio(struct ACEpansionPlugin *plugin, SHORT *leftSample, SHORT *rightSample)
*
* FUNCTION
*   Function to be used when your plugin needs to mix an audio output with the one
*   from ACE.
*   It is called at each PSG emulation step, which means at 125KHz.
*
* INPUTS
*   *leftSample
*     Pointer to store the left audio sample to be mixed with ACE audio.
*   *rightSample
*     Pointer to store the right audio sample to be mixed with ACE audio.
*
* RESULT
*
* SEE ALSO
*
*****************************************************************************
*
*/

VOID GetAudio(UNUSED struct ACEpansionPlugin *plugin, UNUSED SHORT *leftSample, UNUSED SHORT *rightSample)
{
}



/****** printer.acepansion/Printer ******************************************
*
* NAME
*   Printer -- (V1)
*
* SYNOPSIS
*   VOID WritePrinter(struct ACEpansionPlugin *plugin, UBYTE data, BOOL strobe, BOOL *busy)
*
* FUNCTION
*   Function that is called everytime the CPC is accessing the printer port.
*
* INPUTS
*   data
*     8 bits written value (note that CPC is limited to 7 bits values, only Amstrad Plus
*     can actually provide the 8th bit).
*   strobe
*     Status of the /STROBE signal from the printer port.
*   *busy
*     Busy signal value that can be altered.
*
* RESULT
*
* SEE ALSO
*
*****************************************************************************
*
*/

VOID Printer(struct ACEpansionPlugin *plugin, UBYTE data, BOOL strobe, BOOL *busy)
{
    struct PluginData *myPlugin = (struct PluginData *)plugin;

    *busy = FALSE;

    if(myPlugin->printerFile)
    {
        if(myPlugin->previousStrobe && !strobe)
        {
            // Capture printer output to file
            FPutC(myPlugin->printerFile, data);
            Flush(myPlugin->printerFile);
        }
    }

    myPlugin->previousStrobe = strobe;
}



/****** printer.acepansion/Cursor *******************************************
*
* NAME
*   Cursor -- (V1)
*
* SYNOPSIS
*   VOID Cursor(struct ACEpansionPlugin *plugin)
*
* FUNCTION
*   Function which is called everytime the cursor signal from the CRTC is
*   active.
*
* INPUTS
*
* RESULT
*
* SEE ALSO
*
*****************************************************************************
*
*/

VOID Cursor(UNUSED struct ACEpansionPlugin *plugin)
{
}



/****** printer.acepansion/AcknowledgeInterrupt *****************************
*
* NAME
*   AcknowledgeInterrupt -- (V1)
*
* SYNOPSIS
*   VOID AcknowledgeInterrupt(struct ACEpansionPlugin *plugin, UBYTE *ivr)
*
* FUNCTION
*   Function which is called everytime the Z80 is entering an interrupt.
*
*   This happens just before the Z80 is actually entering interrupt routine, and
*   it is the good place to provide support for any expansion supporting
*   interrupt vertorized mode (IM 2).
*
* INPUTS
*   *ivr
*     IVR could be set to the desired value (Interrupt Vector Register).
*
* RESULT
*   ReturnInterrupt()
*
* SEE ALSO
*
*****************************************************************************
*
*/

VOID AcknowledgeInterrupt(UNUSED struct ACEpansionPlugin *plugin, UNUSED UBYTE *ivr)
{
}



/****** printer.library/Joystick ********************************************
*
* NAME
*   Joystick -- (V2)
*
* SYNOPSIS
*   VOID Joystick(struct ACEpansionPlugin *plugin, BOOL com1, BOOL com2, UBYTE *ioData)
*
* FUNCTION
*   Function which is called everytime the CPC is accessing the joystick port.
*   Please note that the joystick port in R/W but you cannot know if the CPC
*   is actually reading or writing to it.
*
* INPUTS
*   com1
*     If com1 is active then the CPC is requesting access to joystick 0.
*   com2
*     If com2 is active then the CPC is requesting access to joystick 1.
*   *ioData
*     If the CPC is reading the joystick port (regular case) ioData will be set
*     to 0xff and you are supposed to set to 0 the bits corresponding to the
*     requested joystick (you should use the ACE_IODATAF_JOYSTICK_XXX defines).
*     Please note the additional ACE_IODATAF_JOYSTICK_PAUSE which is not
*     existing on CPC but that you could use to map the play/pause button of
*     a joystick used in ACE on the 'P' key of the keyboard (or the GX4000
*     pause button).
*     If the CPC is writing the joystick port the ioData will contain the
*     written value to the port and modifying it will have not effect (on a
*     real CPC it could simply destroy the PSG chipset!).
*
* RESULT
*
* SEE ALSO
*
*****************************************************************************
*
*/

VOID Joystick(UNUSED struct ACEpansionPlugin *plugin, UNUSED BOOL com1, UNUSED BOOL com2, UNUSED UBYTE *ioData)
{
}



/****** printer.acepansion/ReturnInterrupt **********************************
*
* NAME
*   ReturnInterrupt -- (V3)
*
* SYNOPSIS
*   VOID ReturnInterrupt(struct ACEpansionPlugin *plugin)
*
* FUNCTION
*   Function which is called everytime the Z80 is returning from an interrupt
*   using RETI instruction.
*
* INPUTS
*
* RESULT
*
* SEE ALSO
*   AcknowledgeInterrupt
*
*****************************************************************************
*
*/

VOID ReturnInterrupt(UNUSED struct ACEpansionPlugin *plugin)
{
}



/****** printer.acepansion/LightDeviceDiodePulse ****************************
*
* NAME
*   LightDeviceDiodePulse -- (V4)
*
* SYNOPSIS
*   VOID LightDeviceDiodePulse(struct ACEpansionPlugin *plugin)
*
* FUNCTION
*   This function is called at each photo-diode pulse, when the beam from
*   the screen monitor is just infront of the light device photo-diode.
*
* INPUTS
*
* RESULT
*
* SEE ALSO
*   LightDeviceButton
*
*****************************************************************************
*
*/

VOID LightDeviceDiodePulse(UNUSED struct ACEpansionPlugin *plugin)
{
}



/****** printer.acepansion/LightDeviceButton ********************************
*
* NAME
*   LightDeviceButton -- (V4)
*
* SYNOPSIS
*   VOID LightDeviceButton(struct ACEpansionPlugin *plugin, BOOL pressed)
*
* FUNCTION
*   This function is called everytime the state of the light device button
*   is changing.
*
* INPUTS
*   pressed
*     New state of the light device button.
*
* RESULT
*
* NOTE
*   When a light device plugin is activated, ACE is mapping this button on
*   mouse select button (LMB), so that on-screen click will emulate light
*   device button.
*
* SEE ALSO
*   LightDeviceDiodePulse
*
*****************************************************************************
*
*/

VOID LightDeviceButton(UNUSED struct ACEpansionPlugin *plugin, UNUSED BOOL pressed)
{
}



/*
** Private printer functions
*/

#ifdef __HAIKU__
static VOID gateHook_SetPrinterFileName(struct PluginData* plugin, const char* fileName)
{
    ObtainSemaphore(plugin->sema);

    FreeVec(plugin->printerFileName);
    plugin->printerFileName = strDup(fileName);

    if(plugin->isActivated)
    {
        if(plugin->printerFile)
        {
            Close(plugin->printerFile);
            plugin->printerFile = (BPTR)0;
        }

        smartOpenPrinterFile(plugin);
    }

    ReleaseSemaphore(plugin->sema);
}
#else
static VOID gateHook_SetPrinterFileName(VOID)
 {
    struct Msg_SetPrinterFileName *msg = (struct Msg_SetPrinterFileName *)REG_A1;

    ObtainSemaphore(msg->plugin->sema);

    FreeVec(msg->plugin->printerFileName);
    msg->plugin->printerFileName = strDup(msg->fileName);

    if(msg->plugin->isActivated)
    {
        if(msg->plugin->printerFile)
        {
            Close(msg->plugin->printerFile);
            msg->plugin->printerFile = (BPTR)0;
        }

        smartOpenPrinterFile(msg->plugin);
    }

    ReleaseSemaphore(msg->plugin->sema);
}
#endif

static BOOL smartOpenPrinterFile(struct PluginData *myPlugin)
{
    if(myPlugin->printerFileName)
    {
        if((myPlugin->printerFile = Open(myPlugin->printerFileName, MODE_READWRITE)))
        {
            Seek(myPlugin->printerFile, 0, OFFSET_END);
            return TRUE;
        }
    }

#ifdef __HAIKU__
	beep();
#else
    DisplayBeep(NULL);
#endif

    return FALSE;
}



/*
** Private helper functions
*/

static ULONG strLen(CONST_STRPTR str)
{
    ULONG count = 0;

    while(*str++ != '\0') count++;

    return count;
}

static STRPTR strDup(CONST_STRPTR str)
{
    STRPTR newStr = NULL;

    if(str)
    {
        ULONG size = strLen(str)+1;
        if((newStr = (STRPTR)AllocVec(size, MEMF_ANY)))
        {
            CopyMem((CONST APTR)str, newStr, size);
        }
    }
    return newStr;
}
